/*-
 * Copyright (C)2014 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2014 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n\
@(#)V.M "VERS;
static const char rcsid[] = "$Id: recobe.c,v 1.2 2014/02/18 16:40:37 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <sys/stat.h>
#include <sysexits.h>
#include <strings.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include <recobe.h>

static void
p256(char *temlate, char from, char to, u_char type, u_char array[256]) {
    int k;
    printf( "static u_char %s%c2%c[] =\n{  "
          , temlate
          , from
          , to
          )
    ;
    for (k = 0; k < 256; ++k) {
        if  (k) {
            if  (!(k & 15)) printf("\n");
            printf(", ");
            if  (!(k & 3)) printf(" ");
        }
        if  (  (type != RECOBE_TEXT)
            || !k
            || array[k]
            ) {
            printf("0x%02X", array[k]);
        } else {
            printf("   0");
    }   }
    printf("\n};\n");
}

static int
test0(char *p) {
    for (; *p; p++) {
        if  (p[1] == ':') printf("%c\t1\n", *(p++));
          else printf("%c\t0\n", *p);
    }
    exit(EX_OK);
}

static int
test1(char *p) {
    int i, j, k;

    for (i = 0; i < recobe_collection.cw; ++i) {
        for (j = 0; j < recobe_collection.cw; ++j) {
            if  (  recobe_collection.nm[i].t
                && (recobe_collection.nm[i].t == recobe_collection.nm[j].t)
                && (!p[1] || (recobe_collection.nm[i].t == (p[1] - '0')))
                ) {
                if  (recobe_collection.ic[j][i]) {
                    for (k = 0; k < 256; ++k) {
                        if  (  (recobe_collection.ic[j][i][recobe_collection.ic[i][j][k]] != k)
                            && (  recobe_collection.ic[j][i][recobe_collection.ic[i][j][k]]
                               || recobe_collection.ic[i][j][k]
                               || (recobe_collection.nm[i].t != RECOBE_TEXT)
                            )  ) {
                            printf( "%c%c[0x%02X]=0x%02X, %c%c[0x%02X]=0x%02X\n"
                                  , recobe_collection.nm[i].n
                                  , recobe_collection.nm[j].n
                                  , k
                                  , recobe_collection.ic[i][j][k]
                                  , recobe_collection.nm[j].n
                                  , recobe_collection.nm[i].n
                                  , recobe_collection.ic[i][j][k]
                                  , recobe_collection.ic[j][i][recobe_collection.ic[i][j][k]]
                                  )
                            ;
                    }   }
                } else if (recobe_collection.ic[i][j]) {
                    u_char o[256];

                    bzero(o, 256);
                    for (k = 0; k < 256; ++k) {
                        o[recobe_collection.ic[i][j][k]] = k;
                    }
                    p256( recobe_collection.p[recobe_collection.nm[j].t]
                        , recobe_collection.nm[j].n
                        , recobe_collection.nm[i].n
                        , recobe_collection.nm[j].t
                        , o
                        )
                    ;
    }   }   }   }
    exit(EX_OK);
}

static int
test2(char *p) {
    int i, j;

    for (i = 0; i < recobe_collection.cw; ++i) if (!p[1] || (i == (p[1] - '0'))) {
        for (j = 0; j < recobe_collection.cw; ++j) if (!p[1] || !p[2] || (j == (p[2] - '0'))) {
            if  (  recobe_collection.ic[i][j]
                && recobe_collection.nm[i].t
                && (recobe_collection.nm[i].t == recobe_collection.nm[j].t)
                ) {
                p256( recobe_collection.p[recobe_collection.nm[i].t]
                    , recobe_collection.nm[i].n
                    , recobe_collection.nm[j].n
                    , recobe_collection.nm[i].t
                    , recobe_collection.ic[i][j]
                    )
                ;
    }   }   }
    exit(EX_OK);
}

static int
test3(char *p) {
    u_char o[256];
    int k;

    for (k = 0; k < 256; ++k) o[k] = k;
    if  (!p[1]) {
        write(fileno(stdout), o, 256);
    } else {
        p256( p[2] ? &p[3] : "X"
            , p[1]
            , p[2] ? p[2] : 'x'
            , RECOBE_TEXT
            , o
            )
        ;
    }
    exit(EX_OK);
}

static void
usage(int ex) {
    int i, j, k;

    fprintf(stderr, "recobe " VERS " @BABOLO " DATE "\nUsage: recobe -c{");
    k = 0;
    for (i = 0; i < recobe_collection.cw; ++i) {
        for (j = 0; j < recobe_collection.cw; ++j) {
            if  (recobe_collection.ic[i][j]) {
                if  (k++) fprintf(stderr, "|");
                fprintf(stderr, "%c%c", recobe_collection.nm[i].n, recobe_collection.nm[j].n);
    }   }   }
    fprintf(stderr, "} [infile]\n");
    for (i = 0; i < recobe_collection.cw; ++i) {
        for (j = 0; j < recobe_collection.cw; ++j) {
            if  (recobe_collection.ic[i][j]) {
                fprintf( stderr
                       , "       -c %c%c       - from %s to %s\n"
                       , recobe_collection.nm[i].n
                       , recobe_collection.nm[j].n
                       , recobe_collection.nm[i].s
                       , recobe_collection.nm[j].s
                       )
                ;
    }   }   }
    exit(ex);
}

static char *
rediropt(babolo_opt **opt, int redirenv) {
    char *o;

    if  ((o = babolo_getoptarg(opt, 0)) && (redirenv > 0)) o = getenv(o);
    return(o);
}

int
main(int argc, char **argv) {
    char *o, *Tflag = NULL, *cflag = NULL;
    int c, i, j, redirenv, ex = EX_OK;
    FILE *in = stdin, *ou = stdout;
    u_char *table = NULL;
    babolo_opt *opt;
    babolo_parm *bp;
    /* ,     flar   '1'.
     * ,        ,
     * . -T,  flar   .
     */
    char *flar = "c:qT:vz:Z:?";

    blin_ctl(BLIN_CTL_FSET, BLIN_VER1);
    opt = NULL;
    if  (babolo_openopt(Bpars_NONU, &opt, argc, argv, flar)) {
        ifBLIN_QW0("openopt");
        ex = EX_SOFTWARE;
        goto out;
    }
    for (redirenv = 0; (c = babolo_getopt(&opt)) != 0; redirenv--) {
        u_int32_t mask;

        mask = 0;
        if  (redirenv < 0) redirenv = 0;
        switch (c) {
        case 'c': if  (!(o = rediropt(&opt, redirenv))) {
                      ifBLIN_QW0("No -c arg");
                      ex = EX_USAGE;
                      goto out;
                  }
                ; cflag = o;
                ; break;
        case 'q': blin_ctl(BLIN_CTL_FRST, BLIN_VERMASK);
                ; break;
        case 'T': if  (!(o = rediropt(&opt, redirenv))) {
                      ifBLIN_QW0("No -T arg");
                      ex = EX_USAGE;
                      goto out;
                  }
                ; Tflag = o;
                ; break;
        case 'v': BLIN_VERBOSE(blin_internal_flags);
                ; break;
        case 'z': mask |= Bpars_NOEN | Bpars_NOAL | Bpars_NONU;
                  /* FALLTHROUGH */
        case 'Z': mask |= Bpars_NOAL | Bpars_NONU;
                ; if  (!(o = rediropt(&opt, redirenv))) {
                      ifBLIN_QW0("getparm -Z arg");
                      ex = EX_SOFTWARE;
                      goto out;
                  }
                ; if  (!(bp = babolo_getparm(mask, (u_char **)&o, (u_char*)"", 0))) {
                      ifBLIN_QW0("No -z arg");
                      ex = EX_USAGE;
                      goto out;
                  }
                ; if  (babolo_openopt(0, &opt, bp->argc, (char **)bp->argv, NULL)) {
                      ifBLIN_QW0("getparm -Z arg");
                      ex = EX_SOFTWARE;
                      goto out;
                  }
                ; break;
        case '?': usage(EX_OK);
        default : ifBLIN_QX0("Illegal flag");
                ; usage(EX_USAGE);
    }   }
    if  (Tflag) {
        switch (Tflag[0]) {
        case '0': ex = test0(flar);  break;
        case '1': ex = test1(Tflag); break;
        case '2': ex = test2(Tflag); break;
        case '3': ex = test3(Tflag); break;
        }
    } else if (cflag) {
        table = NULL;
        for (i = 0; i < recobe_collection.cw; ++i) {
            for (j = 0; j < recobe_collection.cw; ++j) {
                if  (  !cflag[2]
                    && recobe_collection.nm[i].n == cflag[0]
                    && recobe_collection.nm[j].n == cflag[1]
                    ) {
                    table = recobe_collection.ic[i][j];
        }   }   }
        if  (!table) {
            ifBLIN_QX0("No recoding %s", argv[1]);
            ex = EX_USAGE;
            usage(ex);
        }
        if  ((o = babolo_getarg(&opt)) && !(in = fopen(o, "r"))) {
            ifBLIN_QW0("fopen in %s", o);
            ex = EX_IOERR;
            goto out;
        }
        if  ((o = babolo_getarg(&opt)) && !(ou = fopen(o, "w"))) {
            ifBLIN_QW0("fopen ou %s", o);
            ex = EX_IOERR;
            goto out;
        }
        while (0 <= (i = getc(in))) {
            i &= 0xFF;
            j = table[i];
            if  (i & !j) j = '?';
            putc(j, ou);
        }
        fclose(in);
        fclose(ou);
    }
out:
    exit(ex);
}
